@@ -1,5 +1,5 @@
 #
-# Copyright (c) 2022-2023 Andrea Biscuola <a@abiscuola.com>
+# Copyright (c) 2022-2024 Andrea Biscuola <a@abiscuola.com>
 #
 # Permission to use, copy, modify, and distribute this software for any
 # purpose with or without fee is hereby granted, provided that the above
@@ -20,30 +20,30 @@ namespace eval gui {
 	namespace export displaymsg newservertab away closechannel removeuser \
 	    setmode updateuser displaytopic newprivatetab newchanneltab \
 	    privatechat adduser quiet showmode newreadtab init changech \
-	    raisewin onexit savefont newserver
+	    raisewin onexit savefont newserver clearusers setfontcolor \
+	    setchatbackground
 
 	namespace ensemble create
 
-	variable fsett
+	set fsett [dict create]
+	set appsett [dict create]
 
 	set quiet 0
 }
 
 proc ::gui::init {} {
+	variable appsett
+
 	option add *tearOff 0
 
-	wm title . $::progname
-	wm iconname . $::progname
+	wm title . $::progtitle
+	wm iconname . $::progtitle
 
 	bind . <Destroy> {
 		if {"%W" eq "."} {gui onexit}
 	}
 
-	ttk::style theme settings default {
-		ttk::style layout Plain.TNotebook.Tab null
-	}
-
-	font create "$::progname-font" -family TkDefaultFont \
+	font create "$::progname-font" -family $::deffont \
 	    -size 9 -overstrike 0 -underline 0 -weight normal \
 	    -slant roman
 
@@ -59,9 +59,11 @@ proc ::gui::init {} {
 
 	.menubar.edit add command -label "Font" -command "tk fontchooser show"
 	.menubar.edit add command -label "Extensions" -command "::gui::extensions"
+	.menubar.edit add command -label "Themes" -command "::gui::themes"
+	.menubar.edit add separator
+	.menubar.edit add checkbutton -label "Log chats" -onvalue 1 -offvalue 0 \
+	    -variable ::logging -command "::gui::logging"
 
-	.menubar.help add command -label "Wiki" -command "::gui::wiki"
-	.menubar.help add separator
 	.menubar.help add command -label "Website" -command "::gui::website"
 
 	grid [ttk::frame .main] -column 0 -row 0 -sticky nswe -pady 3 \
@@ -75,7 +77,7 @@ proc ::gui::init {} {
 	    -row 0 -columnspan 1 -rowspan 1 -sticky ns
 	.main.servers configure -yscrollcommand ".main.srvscroll set"
 
-	grid [ttk::notebook .main.chatbook -style Plain.TNotebook] \
+	grid [ttk::notebook .main.chatbook] \
 	    -column 2 -row 0 -rowspan 1 -columnspan 1 -sticky nswe
 
 	grid columnconfigure . 0 -weight 1
@@ -87,16 +89,16 @@ proc ::gui::init {} {
 	#
 	# Message levels as shown in the tree:
 	#
-	# - blue: noise.
-	# - orange: user's noise.
-	# - purple: users noise directed at me!
+	# - $::colorinfo: noise.
+	# - $::colormessage: user's noise.
+	# - $::colormention: users noise directed at me!
 	#
 	.main.servers tag configure treenotifymention \
-	    -font "$::progname-font" -foreground "purple"
+	    -font "$::progname-font" -foreground $::colormention
 	.main.servers tag configure treenotifymsg \
-	    -font "$::progname-font" -foreground "orange"
+	    -font "$::progname-font" -foreground $::colormessage
 	.main.servers tag configure treenotifyinfo \
-	    -font "$::progname-font" -foreground "blue"
+	    -font "$::progname-font" -foreground $::colorinfo
 	.main.servers tag configure treenotifynormal \
 	    -font "$::progname-font"
 
@@ -176,11 +178,11 @@ proc ::gui::init {} {
 		set fsett [dict create]
 
 		set font [tk fontchooser configure -font]
-		font configure $::progname-font -family TkDefaultFont \
+		font configure $::progname-font -family $::deffont \
 		    -size 9 -overstrike 0 -underline 0 -weight normal \
 		    -slant roman
 
-		dict set fsett font $::progname-font family TkDefaultFont
+		dict set fsett font $::progname-font family $::deffont
 		dict set fsett font $::progname-font size 9
 		dict set fsett font $::progname-font overstrike 0
 		dict set fsett font $::progname-font underline 0
@@ -224,13 +226,13 @@ proc ::gui::init {} {
 		}
 
 		gui savefont $fsett
+
+		ttk::style configure Treeview -rowheight \
+		    [expr {[font metrics $::progname-font -linespace] + 2}]
 	}
 
-	loadsettings
-}
-
-proc ::gui::wiki {} {
-	plumb {} {} {} $::wiki
+	loadfontsettings
+	loadappsettings
 }
 
 proc ::gui::website {} {
@@ -241,6 +243,19 @@ proc ::gui::onexit {} {
 	irctk onexit
 }
 
+proc ::gui::saveappsettings {} {
+	variable appsett
+
+	file mkdir $::cfgdir
+
+	if {![catch {set f [::open "$::cfgdir/app.conf" w 0640]} errstr]} {
+		puts $f $appsett
+		close $f
+	} else {
+		log console $errstr
+	}
+}
+
 proc ::gui::savefont {sets} {
 	file mkdir $::cfgdir
 
@@ -252,10 +267,60 @@ proc ::gui::savefont {sets} {
 	}
 }
 
+#
+# Load the application settings from file
+#
+proc ::gui::loadappsettings {} {
+	variable appsett
+
+	if {[catch {set fd [::open "$::cfgdir/app.conf" r]} errstr]} {
+		log console $errstr
+
+		source $::defaulttheme
+
+		return
+	}
+
+	set appsett [gets $fd]
+
+	close $fd
+
+	if {[dict exists $appsett logging]} {
+		set ::logging [dict get $appsett logging]
+		log enable $::logging
+	}
+
+	if {[dict exists $appsett themes]} {
+		set themes {}
+
+		foreach file [dict get $appsett themes] {
+			if {[catch {source $file} errstr]} {
+				log console $errstr
+			} else {
+				lappend themes $file
+			}
+		}
+
+		dict set appsett themes $themes
+	}
+
+	if {[dict exists $appsett theme]} {
+		if {[catch {ttk::style theme use [dict get $appsett theme]} errstr]} {
+			log console $errstr
+
+			dict set appsett theme default
+		}
+	}
+
+	source $::defaulttheme
+
+	saveappsettings
+}
+
 #
 # Load the interface settings from file.
 #
-proc ::gui::loadsettings {} {
+proc ::gui::loadfontsettings {} {
 	variable fsett
 
 	if {[catch {set fd [::open "$::cfgdir/font.conf" r]} errstr]} {
@@ -308,36 +373,43 @@ proc ::gui::newserver {{item ""}} {
 	grid [ttk::label .connect.main.netlbl -text "Network"] \
 	    -column 0 -row 0 -columnspan 1 -rowspan 1 -sticky e \
 	    -padx 5 -pady 5
-	grid [ttk::entry .connect.main.net] -column 2 -row 0 \
-	    -columnspan 5 -rowspan 1 -sticky we \
+	grid [ttk::entry .connect.main.net] -column 1 -row 0 \
+	    -columnspan 3 -rowspan 1 -sticky we \
 	    -padx 5 -pady 5
 
 	grid [ttk::label .connect.main.nicklbl -text "Nick"] \
 	    -column 0 -row 2 -columnspan 1 -rowspan 1 -sticky e \
 	    -padx 5 -pady 5
-	grid [ttk::entry .connect.main.nick] -column 2 -row 2 \
+	grid [ttk::entry .connect.main.nick] -column 1 -row 2 \
 	    -columnspan 1 -rowspan 1 -sticky we \
 	    -padx 5 -pady 5
 
 	grid [ttk::label .connect.main.realnlbl -text "Real name"] \
-	    -column 4 -row 2 -columnspan 1 -rowspan 1 -sticky e \
+	    -column 2 -row 2 -columnspan 1 -rowspan 1 -sticky e \
 	    -padx 5 -pady 5
-	grid [ttk::entry .connect.main.realn] -column 6 -row 2 \
+	grid [ttk::entry .connect.main.realn] -column 3 -row 2 \
+	    -columnspan 1 -rowspan 1 -sticky we \
+	    -padx 5 -pady 5
+
+	grid [ttk::label .connect.main.userlbl -text "User"] \
+	    -column 0 -row 3 -columnspan 1 -rowspan 1 -sticky e \
+	    -padx 5 -pady 5
+	grid [ttk::entry .connect.main.user] -column 1 -row 3 \
 	    -columnspan 1 -rowspan 1 -sticky we \
 	    -padx 5 -pady 5
 
 	grid [ttk::label .connect.main.passlbl -text "Password"] \
-	    -column 0 -row 4 -columnspan 1 -rowspan 1 -sticky e \
+	    -column 2 -row 3 -columnspan 1 -rowspan 1 -sticky e \
 	    -padx 5 -pady 5
-	grid [ttk::entry .connect.main.pass -show "*"] -column 2 -row 4 \
+	grid [ttk::entry .connect.main.pass -show "*"] -column 3 -row 3 \
 	    -columnspan 1 -rowspan 1 -sticky we \
 	    -padx 5 -pady 5
 
 	grid [ttk::label .connect.main.cmdlbl -text "Connection string"] \
-	    -column 0 -row 6 -columnspan 1 -rowspan 1 \
+	    -column 0 -row 4 -columnspan 1 -rowspan 1 \
 	    -padx 5 -pady 5 -sticky e
 	grid [ttk::entry .connect.main.cmd] \
-	    -column 2 -row 6 -columnspan 5 -rowspan 1 \
+	    -column 1 -row 4 -columnspan 3 -rowspan 1 \
 	    -padx 5 -pady 5 -sticky we
 
 	if {$item eq ""} {
@@ -354,6 +426,7 @@ proc ::gui::newserver {{item ""}} {
 		.connect.main.realn insert 0 [lindex $net 2]
 		.connect.main.pass insert 0 [lindex $net 3]
 		.connect.main.cmd insert 0 [lindex $net 4]
+		.connect.main.user insert 0 [lindex $net 5]
 
 		grid [ttk::button .connect.bottom.ok -text "OK" \
 		    -command "::gui::modify"] \
@@ -370,17 +443,18 @@ proc ::gui::newserver {{item ""}} {
 	grid columnconfigure .connect.main 0 -weight 1
 	grid rowconfigure .connect.main 0 -weight 1
 
+	grid columnconfigure .connect.main 1 -weight 1
 	grid columnconfigure .connect.main 2 -weight 1
-	grid columnconfigure .connect.main 4 -weight 1
-	grid columnconfigure .connect.main 6 -weight 1
+	grid columnconfigure .connect.main 3 -weight 1
 
+	grid rowconfigure .connect.main 1 -weight 1
 	grid rowconfigure .connect.main 2 -weight 1
-	grid rowconfigure .connect.main 4 -weight 1
-	grid rowconfigure .connect.main 6 -weight 1
+	grid rowconfigure .connect.main 3 -weight 1
 }
 
-proc ::gui::validform {net nick name cmd} {
+proc ::gui::validform {net user nick name cmd} {
 	if {![regexp {[[:alnum:]]+} $net]} {return 0}
+	if {[regexp {[[:space:]]+} $user]} {return 0}
 	if {"$nick" eq "" || [regexp {[[:space:]]+} $nick]} {return 0}
 	if {![regexp {[:alnum:]} $name]} {return 0}
 	if {"$cmd" eq ""} {return 0}
@@ -390,15 +464,17 @@ proc ::gui::validform {net nick name cmd} {
 
 proc ::gui::connect {} {
 	set net [.connect.main.net get]
+	set user [.connect.main.user get]
 	set nick [.connect.main.nick get]
 	set name [.connect.main.realn get]
 	set pass [.connect.main.pass get]
 	set cmd [.connect.main.cmd get]
 
-	if {![validform "$net" "$nick" "$name" "$cmd"]} {
+	if {![validform "$net" "$user" "$nick" "$name" "$cmd"]} {
 		set str [format "%s\n\n\n%s\n\n%s\n\n%s\n\n%s" \
-		    "One ore more values are not valid." \
+		    "One or more values are not valid." \
 		    "* The network name must contain alphanumeric characters." \
+		    "* The user can be empty or must contain alphanumeric characters." \
 		    "* The nick can not be empty, or contain spaces." \
 		    "* The real name must contain alphanumeric characters." \
 		    "* The command to run can not be empty."]
@@ -409,7 +485,7 @@ proc ::gui::connect {} {
 		return
 	}
 
-	set fd [irctk connect $net $nick $name $pass "$cmd"]
+	set fd [irctk connect $net $user $nick $name $pass "$cmd"]
 
 	destroy .connect
 }
@@ -417,14 +493,16 @@ proc ::gui::connect {} {
 proc ::gui::modify {} {
 	set net [.connect.main.net get]
 	set nick [.connect.main.nick get]
+	set user [.connect.main.user get]
 	set name [.connect.main.realn get]
 	set pass [.connect.main.pass get]
 	set cmd [.connect.main.cmd get]
 
-	if {![validform "$net" "$nick" "$name" "$cmd"]} {
+	if {![validform "$net" "$user" "$nick" "$name" "$cmd"]} {
 		set str [format "%s\n\n\n%s\n\n%s\n\n%s\n\n%s" \
-		    "One ore more values are not valid." \
+		    "One or more values are not valid." \
 		    "* The network name must contain alphanumeric characters." \
+		    "* The user can be empty or must contain alphanumeric characters." \
 		    "* The nick can not be empty, or contain spaces." \
 		    "* The real name must contain alphanumeric characters." \
 		    "* The command to run can not be empty."]
@@ -437,12 +515,13 @@ proc ::gui::modify {} {
 
 	destroy .connect
 
-	irctk save $net $nick "$name" "$pass" "$cmd"
+	irctk save $net $user $nick "$name" "$pass" "$cmd"
 }
 
 proc ::gui::newservertab {fd nick network id} {
 	if {![winfo exists .main.chatbook.$id]} {
-		.main.chatbook add [ttk::frame .main.chatbook.$id] -sticky nswe
+		.main.chatbook add [ttk::frame .main.chatbook.$id] \
+		    -text "$network" -sticky nswe
 
 		grid [ttk::frame .main.chatbook.$id.top] -column 0 -row 0 \
 		    -columnspan 1 -rowspan 1 -sticky nswe -padx 3
@@ -451,8 +530,9 @@ proc ::gui::newservertab {fd nick network id} {
 
 		grid [text .main.chatbook.$id.top.chat -state disabled \
 		    -font $::progname-font -exportselection 1 -cursor top_left_arrow \
-		    -spacing3 5px] -column 0 -row 0 -sticky nswe -columnspan 1 \
+		    -spacing3 5] -column 0 -row 0 -sticky nswe -columnspan 1 \
 		    -rowspan 1
+		.main.chatbook.$id.top.chat tag configure default -foreground $::colordefault
 		.main.chatbook.$id.top.chat tag bind sel <ButtonRelease-3> \
 		    "::gui::textclick $id \"$network\" \"$network\" .main.chatbook.$id.top.chat"
 
@@ -489,7 +569,8 @@ proc ::gui::newservertab {fd nick network id} {
 
 proc ::gui::newchanneltab {fd nick parent id channel network} {
 	if {![winfo exists .main.chatbook.$id]} {
-		.main.chatbook add [ttk::frame .main.chatbook.$id] -sticky nswe
+		.main.chatbook add [ttk::frame .main.chatbook.$id] \
+		    -text "$channel" -sticky nswe
 
 		grid [ttk::frame .main.chatbook.$id.top] -column 0 -row 0 \
 		    -columnspan 1 -rowspan 1 -sticky nswe -padx 3
@@ -502,11 +583,12 @@ proc ::gui::newchanneltab {fd nick parent id channel network} {
 
 		grid [text .main.chatbook.$id.top.chat -state disabled \
 		    -font $::progname-font -exportselection 1 -cursor top_left_arrow \
-		    -spacing3 5px] -column 0 -row 1 -sticky nswe -columnspan 1 \
+		    -spacing3 5] -column 0 -row 1 -sticky nswe -columnspan 1 \
 		    -rowspan 1
 		.main.chatbook.$id.top.chat tag bind sel <ButtonRelease-3> \
 		    "::gui::textclick $id \"$network\" $channel .main.chatbook.$id.top.chat"
-		.main.chatbook.$id.top.chat tag configure mention -foreground purple
+		.main.chatbook.$id.top.chat tag configure default -foreground $::colordefault
+		.main.chatbook.$id.top.chat tag configure mention -foreground $::colormention
 
 		grid [ttk::label .main.chatbook.$id.bottom.nick -text $nick \
 		    -font $::progname-font] -column 0 -row 0 -columnspan 1 \
@@ -526,15 +608,15 @@ proc ::gui::newchanneltab {fd nick parent id channel network} {
 		    -columnspan 1 -rowspan 2
 
 		.main.chatbook.$id.users tag configure founder \
-		    -font "$::progname-font" -foreground "red"
+		    -font "$::progname-font" -foreground $::colorfounder
 		.main.chatbook.$id.users tag configure protected \
-		    -font "$::progname-font" -foreground "purple"
+		    -font "$::progname-font" -foreground $::colorprotected
 		.main.chatbook.$id.users tag configure operator \
-		    -font "$::progname-font" -foreground "green"
+		    -font "$::progname-font" -foreground $::coloroperator
 		.main.chatbook.$id.users tag configure halfop \
-		    -font "$::progname-font" -foreground "orange"
+		    -font "$::progname-font" -foreground $::colorhalfop
 		.main.chatbook.$id.users tag configure voice \
-		    -font "$::progname-font" -foreground "blue"
+		    -font "$::progname-font" -foreground $::colorvoice
 		.main.chatbook.$id.users tag configure user \
 		    -font "$::progname-font"
 
@@ -570,16 +652,10 @@ proc ::gui::newchanneltab {fd nick parent id channel network} {
 	    "gui privatechat $fd $id"
 }
 
-proc ::gui::privatechat {fd id} {
-	set sl [.main.chatbook.$id.users selection]
-	if {[llength $sl] == 0} {return}
-
-	irctk privatechan $fd [lindex $sl 0]
-}
-
-proc ::gui::newprivatetab {fd nick parent id channel network} {
+proc ::gui::newprivatetab {fd nick parent id channel network mode} {
 	if {![winfo exists .main.chatbook.$id]} {
-		.main.chatbook add [ttk::frame .main.chatbook.$id] -sticky nswe
+		.main.chatbook add [ttk::frame .main.chatbook.$id] \
+		    -text "$channel" -sticky nswe
 
 		grid [ttk::frame .main.chatbook.$id.top] -column 0 -row 0 \
 		    -columnspan 1 -rowspan 1 -sticky nswe -padx 3
@@ -588,8 +664,9 @@ proc ::gui::newprivatetab {fd nick parent id channel network} {
 
 		grid [text .main.chatbook.$id.top.chat -state disabled \
 		    -font $::progname-font -exportselection 1 -cursor top_left_arrow \
-		    -spacing3 5px] -column 0 -row 0 -sticky nswe -columnspan 1 \
+		    -spacing3 5] -column 0 -row 0 -sticky nswe -columnspan 1 \
 		    -rowspan 1
+		.main.chatbook.$id.top.chat tag configure default -foreground $::colordefault
 		.main.chatbook.$id.top.chat tag bind sel <ButtonRelease-3> \
 		    "::gui::textclick $id \"$network\" $channel .main.chatbook.$id.top.chat"
 
@@ -617,8 +694,10 @@ proc ::gui::newprivatetab {fd nick parent id channel network} {
 		grid columnconfigure .main.chatbook.$id.bottom 1 -weight 1
 
 		.main.servers insert $parent end -id $id -text $channel
-		.main.servers see $id
 		.main.servers tag add treenotifynormal $id
+		if {$mode eq "focus"} {
+			.main.servers see $id
+		}
 	} else {
 		bind .main.chatbook.$id.bottom.message <Return> ""
 	}
@@ -628,20 +707,23 @@ proc ::gui::newprivatetab {fd nick parent id channel network} {
 }
 
 proc ::gui::newreadtab {parent id channel network} {
-	.main.chatbook add [ttk::frame .main.chatbook.$id] -sticky nswe
+	.main.chatbook add [ttk::frame .main.chatbook.$id] \
+	    -text "$channel" -sticky nswe
 
 	grid [ttk::frame .main.chatbook.$id.top] -column 0 -row 0 \
 	    -columnspan 1 -rowspan 1 -sticky nswe -padx 3
 
 	grid [text .main.chatbook.$id.top.chat -state disabled \
 	    -font $::progname-font -exportselection 1 -cursor top_left_arrow \
-	    -spacing3 5px] -column 0 -row 0 -sticky nswe -columnspan 1 -rowspan 1
+	    -spacing3 5] -column 0 -row 0 -sticky nswe -columnspan 1 -rowspan 1
+	.main.chatbook.$id.top.chat tag configure default -foreground $::colordefault
 	.main.chatbook.$id.top.chat tag bind sel <ButtonRelease-3> \
 	    "::gui::textclick $id \"$network\" $channel .main.chatbook.$id.top.chat"
 
 	grid [ttk::scrollbar .main.chatbook.$id.top.textscroll -orient vertical \
 	    -command ".main.chatbook.$id.top.chat yview"] -column 1 \
 	    -row 0 -columnspan 1 -rowspan 1 -sticky ns
+
 	.main.chatbook.$id.top.chat configure -yscrollcommand \
 	    ".main.chatbook.$id.top.textscroll set"
 
@@ -657,6 +739,13 @@ proc ::gui::newreadtab {parent id channel network} {
 	.main.servers tag add treenotifynormal $id
 }
 
+proc ::gui::privatechat {fd id} {
+	set sl [.main.chatbook.$id.users selection]
+	if {[llength $sl] == 0} {return}
+
+	irctk privatechan $fd [lindex $sl 0]
+}
+
 #
 # Are we sleeping?
 #
@@ -716,11 +805,10 @@ proc ::gui::displaymsg {msg nick id level datetime} {
 		#
 		if {$level eq "mention"} {
 			.main.chatbook.$id.top.chat insert end \
-			    "$datetime $nick:\t$msg\n" \
-			mention
+			    "$datetime $nick:\t$msg\n" mention
 		} else {
 			.main.chatbook.$id.top.chat insert end \
-			    "$datetime $nick:\t$msg\n"
+			    "$datetime $nick:\t$msg\n" default
 		}
 
 		.main.chatbook.$id.top.chat configure -state disabled
@@ -794,7 +882,7 @@ proc ::gui::adduser {id nick} {
 	if {![winfo exists .main.chatbook.$id.users]} {return}
 
 	set user [string trimleft $nick "~&@%+"]
-	set literal [regsub -all -- {\\} $user {\\\\}]
+	set literal [regsub -all -- {([^0-9a-zA-Z])} $user {\\\1}]
 
 	if {[.main.chatbook.$id.users exists $user]} {return}
 
@@ -813,6 +901,14 @@ proc ::gui::adduser {id nick} {
 	}
 }
 
+proc ::gui::clearusers {id} {
+	set tree .main.chatbook.$id.users
+
+	foreach item [$tree children {}] {
+		$tree delete [escapestr $item]
+	}
+}
+
 proc ::gui::escapestr {text} {
 	regsub -all -- {\\} $text {\\\\}
 }
@@ -956,9 +1052,10 @@ proc ::gui::extensions {} {
 	grid [ttk::treeview .extensions.top.view] -column 0 -row 0 \
 	    -columnspan 3 -rowspan 1 -sticky nswe
 
-	.extensions.top.view configure -columns {Version Path}
+	.extensions.top.view configure -columns {Version Capabilities Path}
 	.extensions.top.view heading {#0} -text "Name"
 	.extensions.top.view heading {Version} -text "Version"
+	.extensions.top.view heading {Capabilities} -text "Capabilities"
 	.extensions.top.view heading {Path} -text "Path"
 
 	grid [ttk::button .extensions.bottom.close -text "Close" \
@@ -982,7 +1079,8 @@ proc ::gui::extensions {} {
 
 		.extensions.top.view insert {} end -id [lindex $ext 0] -text $id
 		.extensions.top.view set $id Version [lindex $ext 1]
-		.extensions.top.view set $id Path [lindex $ext 2]
+		.extensions.top.view set $id Capabilities [concat [lindex $ext 2]]
+		.extensions.top.view set $id Path [lindex $ext 3]
 	}
 }
 
@@ -1002,6 +1100,7 @@ proc ::gui::isloaded {msg} {
 	.extensions.top.view insert {} end -id $id \
 	    -text $id
 	.extensions.top.view set $id Version [dict get $msg version]
+	.extensions.top.view set $id Capabilities [dict get $msg caps]
 	.extensions.top.view set $id Path [dict get $msg path]
 
 	exts load [dict get $msg path] "::irctk::extmsg"
@@ -1057,3 +1156,191 @@ proc ::gui::changech {{id 0}} {
 	.main.servers tag remove treenotifymsg $slt
 	.main.servers tag remove treenotifymention $slt
 }
+
+proc ::gui::logging {} {
+	variable appsett
+
+	log enable $::logging
+
+	dict set appsett logging $::logging
+	saveappsettings
+}
+
+proc ::gui::themes {} {
+	variable appsett
+
+	if {[winfo exists .themes]} {return}
+
+	tk::toplevel .themes
+
+	wm title .themes "Themes"
+
+	grid [ttk::frame .themes.top -padding "3 3 12 12"] -column 0 \
+	    -row 0 -sticky nswe
+	grid [ttk::frame .themes.bottom -padding "3 3 12 12"] -column 0 \
+	    -row 1 -sticky nswe
+
+	grid [ttk::treeview .themes.top.view -selectmode browse] \
+	    -column 0 -row 0 -columnspan 3 -rowspan 1 -sticky nswe
+	grid [ttk::treeview .themes.top.files] -column 0 -row 1 \
+	    -columnspan 3 -rowspan 1 -sticky nswe
+
+	.themes.top.view heading {#0} -text "Name"
+	.themes.top.files heading {#0} -text "Loaded theme files"
+
+	grid [ttk::button .themes.bottom.close -text "Close" \
+	    -command "destroy .themes"] -row 0 -column 0 \
+	    -sticky ne -padx 20
+	grid [ttk::button .themes.bottom.load -text "Load" \
+	    -command "::gui::loadtheme"] -row 0 -column 1 \
+	    -sticky ne -padx 20
+	grid [ttk::button .themes.bottom.unload -text "Remove" \
+	    -command "::gui::removetheme"] -row 0 -column 2 \
+	    -sticky ne -padx 20
+
+	grid columnconfigure .themes 0 -weight 1
+	grid rowconfigure .themes 0 -weight 1
+
+	grid columnconfigure .themes.top 0 -weight 1
+	grid rowconfigure .themes.top 0 -weight 1
+	grid columnconfigure .themes.top 0 -weight 1
+	grid rowconfigure .themes.top 1 -weight 1
+
+	bind .themes.top.view <<TreeviewSelect>> {
+		ttk::style theme use [.themes.top.view selection]
+		source $::defaulttheme
+
+		dict set ::gui::appsett theme [ttk::style theme use]
+
+		::gui::saveappsettings
+	}
+
+	listthemes
+}
+
+proc ::gui::listthemes {} {
+	variable appsett
+
+	foreach item [.themes.top.view children {}] {
+		.themes.top.view delete $item
+	}
+	foreach item [.themes.top.files children {}] {
+		.themes.top.files delete $item
+	}
+
+	foreach theme [ttk::style theme names] {
+		.themes.top.view insert {} end -id $theme -text $theme
+	}
+	if {[dict exists $appsett themes]} {
+		foreach file [dict get $appsett themes] {
+			.themes.top.files insert {} end -id $file -text $file
+		}
+	}
+
+	.themes.top.view selection set [ttk::style theme use]
+}
+
+proc ::gui::loadtheme {} {
+	variable appsett
+
+	set path [tk_getOpenFile]
+
+	if {"$path" eq ""} {return}
+
+	if {[catch {source $path} errstr]} {
+		log console $errstr
+	}
+
+	if {![dict exists $appsett themes]} {
+		dict set appsett themes {}
+	}
+
+	if {[lsearch -exact [dict get $appsett themes] $path] == -1} {
+		set themes [dict get $appsett themes]
+		lappend themes $path
+		dict set appsett themes $themes
+		saveappsettings
+	}
+
+	listthemes
+}
+
+proc ::gui::removetheme {} {
+	variable appsett
+
+	set themes [dict get $appsett themes]
+
+	foreach file [.themes.top.files selection] {
+		set id [lsearch -exact $themes $file]
+
+		if {$id == -1} {continue}
+
+		set themes [lreplace $themes $id $id]
+	}
+
+	dict set appsett themes $themes
+
+	saveappsettings
+
+	listthemes
+}
+
+proc ::gui::setfontcolor {tag color} {
+	switch -exact $tag {
+		mention {
+			set ::colormention $color
+			.main.servers tag configure treenotifymention \
+			    -foreground $color
+			settextchattag $tag $color
+		} message {
+			set ::colormessage $color
+			.main.servers tag configure treenotifymsg \
+			    -foreground $color
+		} info {
+			set ::colorinfo $color
+			.main.servers tag configure treenotifyinfo \
+			    -foreground $color
+		} founder {
+			set ::colorfounder $color
+			setuserstag $tag $color
+		} protected {
+			set ::colorfounder $color
+			setuserstag $tag $color
+		} operator {
+			set ::colorfounder $color
+			setuserstag $tag $color
+		} halfop {
+			set ::colorfounder $color
+			setuserstag $tag $color
+		} voice {
+			set ::colorfounder $color
+			setuserstag $tag $color
+		} default {
+			set ::colordefault $color
+			.main.servers tag configure treenotifynormal \
+			    -foreground $color
+			settextchattag default $color
+			setuserstag user $color
+		}
+	}
+}
+
+proc ::gui::setchatbackground {color} {
+	foreach tab [winfo children .main.chatbook] {
+		$tab.top.chat configure -background $color
+	}
+}
+
+proc ::gui::settextchattag {tag color} {
+	foreach tab [winfo children .main.chatbook] {
+		$tab.top.chat tag configure $tag -foreground $color
+	}
+}
+
+proc ::gui::setuserstag {tag color} {
+	foreach tab [winfo children .main.chatbook] {
+		if {[winfo exists $tab.users]} {
+			$tab.users tag configure $tag -foreground $color
+		}
+	}
+}
