@@ -1,5 +1,5 @@
 #
-# Copyright (c) 2022-2023 Andrea Biscuola <a@abiscuola.com>
+# Copyright (c) 2022-2024 Andrea Biscuola <a@abiscuola.com>
 # Copyright (c) 2023 Omar Polo <op@omarpolo.com>
 #
 # Permission to use, copy, modify, and distribute this software for any
@@ -32,7 +32,7 @@ proc ::irctk::init {} {
 	variable servers
 	variable newservers
 
-	set nextid 1
+	set nextid 0
 
 	# Dictionary with all the servers we are connected to
 	set servers [dict create]
@@ -55,8 +55,9 @@ proc ::irctk::init {} {
 			set name [lindex $cfg 2]
 			set pass [lindex $cfg 3]
 			set cmd [lindex $cfg 4]
+			set user [lindex $cfg 5]
 
-			connect $network $nick $name $pass "$cmd"
+			connect $network "$user" $nick $name "$pass" "$cmd"
 
 			if {[catch {close $sfd} errstr]} {
 				log console "::irctk::init: $errstr"
@@ -84,37 +85,25 @@ proc ::irctk::onexit {} {
 proc ::irctk::genid {} {
 	variable nextid
 
-	set nextid [expr $nextid + 1]
-
-	expr "$nextid - 1"
+	incr nextid
 }
 
 proc ::irctk::autojoin {fd} {
 	variable servers
 
-	#
-	# For every channel automatically saved for a network,
-	# send a "join" command, randomizing the time between
-	# the various requests of a value comprised between 500
-	# and 2000 milliseconds.
-	#
-	# Some servers are susceptible and think we want to do
-	# "bad" things to them. This way we make them happy.
-	#
+	set channels {}
+
 	foreach ch  [dict get $servers $fd autojoin] {
 		switch -regexp $ch {
 			server {}
-			^\[&\#!\+\] {
-				after [expr int(500 + rand() * 2000)] \
-				    "irc join $fd {} $ch"
-			} default {
-				privatechan $fd $ch
-			}
+			default {lappend channels $ch}
 		}
 	}
+
+	if {[llength $channels] > 0} {irc join $fd {} [join $channels ","]}
 }
 
-proc ::irctk::connect {network nick name pass cmd {oldch none}} {
+proc ::irctk::connect {network user nick name pass cmd {oldch none}} {
 	variable servers
 	variable fd
 
@@ -178,13 +167,20 @@ proc ::irctk::connect {network nick name pass cmd {oldch none}} {
 	gui displaymsg "Connecting to $network with \"$cmd\"" "*" \
 	    $sid info $date
 
+	#
+	# We use the logged in user by default if none is passed
+	#
+	if {"$user" eq ""} {set user $::env(USER)}
+
 	dict set servers $fd network $network
+	dict set servers $fd user "$user"
 	dict set servers $fd nick $nick
 	dict set servers $fd name $name
-	dict set servers $fd pass $pass
+	dict set servers $fd pass "$pass"
 	dict set servers $fd cmd "$cmd"
 	dict set servers $fd away 0
 	dict set servers $fd connected 1
+	dict set servers $fd pendingcaps 0
 
 	#
 	# Request IRCv3 capabilities list.
@@ -196,48 +192,7 @@ proc ::irctk::connect {network nick name pass cmd {oldch none}} {
 	#
 	# Register the user for the connection.
 	#
-	irc register $fd {} $::env(USER) $nick $pass 0 $name
-
-	#
-	# Send to the extensions the registration transaction
-	#
-	if {"$pass" ne ""} {
-		set regmsg [exts newmsg]
-
-		dict set regmsg type irc
-		dict set regmsg network $network
-		dict set regmsg nick $nick
-		dict set regmsg channel $network
-		dict set regmsg cid [dict get $servers $fd channels server id]
-		dict set regmsg command pass
-		dict set regmsg args [list $pass]
-
-		exts writemsg $regmsg
-	}
-
-	set regmsg [exts newmsg]
-
-	dict set regmsg type irc
-	dict set regmsg network $network
-	dict set regmsg nick $nick
-	dict set regmsg channel $network
-	dict set regmsg cid [dict get $servers $fd channels server id]
-	dict set regmsg command nick
-	dict set regmsg args [list $nick]
-
-	exts writemsg $regmsg
-
-	set regmsg [exts newmsg]
-
-	dict set regmsg type irc
-	dict set regmsg network $network
-	dict set regmsg nick $nick
-	dict set regmsg channel $network
-	dict set regmsg cid [dict get $servers $fd channels server id]
-	dict set regmsg command "user"
-	dict set regmsg args [list $::env(USER) 0 "*" "$name"]
-
-	exts writemsg $regmsg
+	irc register $fd {} "$user" $nick "$pass" 0 $name
 
 	set chans ""
 
@@ -290,7 +245,7 @@ proc ::irctk::connect {network nick name pass cmd {oldch none}} {
 			log console "::irctk::connect: $errstr"
 		}
 
-		set params [::list $network $nick $name $pass "$cmd"]
+		set params [::list $network $nick $name "$pass" "$cmd" "$user"]
 		set cfgdir "$::cfgdir/auto/$network"
 
 		if {[catch {file mkdir $cfgdir} errstr]} {
@@ -328,7 +283,9 @@ proc ::irctk::reconnect {chan} {
 	variable servers
 	variable newservers
 
-	if {![dict get $servers $chan connected]} {return}
+	if {[dict exists $servers $chan connected]} {
+		if {![dict get $servers $chan connected]} {return}
+	}
 
 	#
 	# Remove outstanding PING and PONG events, if any.
@@ -360,11 +317,13 @@ proc ::irctk::reconnect {chan} {
 	#
 	if {![dict exists $newservers network $network]} {
 		set nick [dict get $servers $chan nick]
+		set user [dict get $servers $chan user]
 		set name [dict get $servers $chan name]
 		set pass [dict get $servers $chan pass]
 		set cmd [dict get $servers $chan cmd]
 	} else {
 		set nick [dict get $newservers network $network nick]
+		set user [dict get $newservers network $network user]
 		set name [dict get $newservers network $network name]
 		set pass [dict get $newservers network $network pass]
 		set cmd [dict get $newservers network $network cmd]
@@ -388,7 +347,7 @@ proc ::irctk::reconnect {chan} {
 	# We do not want to be banned, right? RIGHT???!?
 	#
 	dict set servers $chan retry [after $retry "irctk connect \
-	    \"$network\" \"$nick\" \"$name\" \"$pass\" \"$cmd\" $chan"]
+	    \"$network\" \"$user\" \"$nick\" \"$name\" \"$pass\" \"$cmd\" $chan"]
 }
 
 #
@@ -401,6 +360,10 @@ proc ::irctk::killconnection {id} {
 	reconnect $chan
 }
 
+proc hascap {caps cap} {
+	expr {[lsearch -nocase -bisect $caps $cap] != -1}
+}
+
 proc ::irctk::readmsg {chan} {
 	variable servers
 
@@ -460,19 +423,34 @@ proc ::irctk::readmsg {chan} {
 		cap {
 			set recmd [lindex [dict get $rmsg args] 1]
 
-			switch -nocase -- $recmd {
+			switch -nocase -regexp -- $recmd {
 				ls {
-					irc cap $chan {} :server-time
-				} ack {
-					# XXX: this assumes we'll only
-					# get _one_ ACK or NAK, since
-					# we only send _one_ CAP REQ.
-					# Revisit if we want to start
-					# issuing more than one CAP
-					# REQ.
-					irc cap $chan {} end
-				} nak {
-					irc cap $chan {} end
+					# XXX: we should support CAP LS 302 with
+					# multiple lines of response eventually
+					# and also caps with arguments.
+
+					set caps [lrange [dict get $rmsg args] 2 end]
+					set caps [string trimleft "$caps" :]
+					set caps [lsort -nocase $caps]
+					dict set $servers $chan caps $caps
+
+					set p [dict get $servers $chan pendingcaps]
+
+					if {[hascap $caps server-time]} {
+						irc cap $chan {} server-time
+						incr p
+					}
+
+					dict set servers $chan pendingcaps $p
+					if {$p == 0} {
+						irc cap $chan {} end
+					}
+				} ack|nak {
+					set n [dict get $servers $chan pendingcaps]
+					dict set servers $chan pendingcaps [incr n -1]
+					if {$n == 0} {
+						irc cap $chan {} end
+					}
 				}
 			}
 		} privmsg|notice {
@@ -495,11 +473,6 @@ proc ::irctk::readmsg {chan} {
 			log write [dict get $servers $chan channels server log] \
 			    * [::join [dict get $rmsg args] " "] [dict get $rmsg date]
 
-			gui closechannel [dict get $servers $chan \
-			    channels server id]
-
-			quitserver $chan
-
 			return
 		} mode {
 			managemode $chan
@@ -521,7 +494,7 @@ proc ::irctk::readmsg {chan} {
 	if {![dict exists $rmsg user]} {
 		dict set rmsg nick ""
 	} elseif {![dict exists $rmsg nick]} {
-		dict set rmsg nick [dict get $rmsg user]]
+		dict set rmsg nick [dict get $rmsg user]
 	}
 
 	if {"[focus]" eq ""} {
@@ -682,7 +655,8 @@ proc ::irctk::managenick {chan} {
 		set name [dict get $servers $chan name]
 		set pass [dict get $servers $chan pass]
 		set cmd [dict get $servers $chan cmd]
-		set params [::list $network $newnick $name $pass "$cmd"]
+		set user [dict get $servers $chan user]
+		set params [::list $network $newnick $name "$pass" "$cmd" "$user"]
 		set cfgdir "$::cfgdir/auto/$network"
 
 		if {[catch {file mkdir $cfgdir} errstr]} {
@@ -789,12 +763,14 @@ proc ::irctk::privmsg {chan} {
 			privatechan $chan [dict get $msg nick]
 
 			set dest [dict get $msg nick]
-		} elseif {![dict exists $servers $chan channels $target]} {
-			set dest server
 		} else {
 			set dest $target
 		}
 
+		if {![dict exists $servers $chan channels $target]} {
+			privatechan $chan $target nofocus
+		}
+
 		if {":ACTION" eq [lindex [dict get $msg args] 0]} {
 			dict set msg args [lreplace [dict get $msg args] \
 			    0 0 [dict get $msg nick]]
@@ -840,7 +816,7 @@ proc ::irctk::privmsg {chan} {
 	}
 }
 
-proc ::irctk::privatechan {chan nick} {
+proc ::irctk::privatechan {chan nick {mode focus}} {
 	variable servers
 
 	#
@@ -862,7 +838,7 @@ proc ::irctk::privatechan {chan nick} {
 	set id [dict get $servers $chan channels $nick id]
 
 	gui newprivatetab $chan [dict get $servers $chan nick] \
-	    $parent $id $nick [dict get $servers $chan network]
+	    $parent $id $nick [dict get $servers $chan network] $mode
 }
 
 #
@@ -893,6 +869,7 @@ proc ::irctk::joinch {chan} {
 		dict set servers $chan channels $channel log $fd
 		dict set servers $chan channels $channel id [genid]
 		dict set servers $chan channels $channel readonly 0
+		dict set servers $chan channels $channel names 0
 
 		if {![catch {set fd [open $::cfgdir/auto/$network/join.conf w 0640]} errstr]} {
 			dict for {ch log} [dict get $servers $chan channels] {
@@ -936,6 +913,10 @@ proc ::irctk::findchan {network {channel ""}} {
 	variable servers
 
 	dict for {chan net} $servers {
+		if {![dict exists $net network]} {
+			continue
+		}
+
 		if {[dict get $net network] ne $network} {
 			continue
 		} elseif {"$channel" eq "[dict get $net network]"} {
@@ -1086,15 +1067,25 @@ proc ::irctk::writemsg {chan id tags text {loop 0}} {
 		dict set emsg command privmsg
 		dict set emsg args $text
 
-		irc privmsg $chan {} $channel $text
+		switch -exact -- [irc privmsg $chan {} $channel $text] {
+			-2 {
+				gui displaymsg "message too long" * $id none $date
 
-		gui displaymsg $text $nick $id none $date
+				return
+			} -1 {
+				gui displaymsg "can not send message" * $id none $date
 
-		log write $lfd $nick $text $tstamp
+				return
+			} default {
+				gui displaymsg $text $nick $id none $date
 
-		exts writemsg $emsg
+				log write $lfd $nick $text $tstamp
 
-		return
+				exts writemsg $emsg
+
+				return
+			}
+		}
 	}
 
 	set args [split [string trimleft $text "/"] " "]
@@ -1148,7 +1139,15 @@ proc ::irctk::writemsg {chan id tags text {loop 0}} {
 
 			prepareme $nick $id $channel
 		} quit {
-			irc quit $chan {} [::join $args " "]
+			switch -exact -- [irc quit $chan {} [::join $args " "]] {
+				-2 {
+					gui displaymsg "message too long" \
+					    * $id none $date
+				} -1 {
+					gui displaymsg "can not send message" \
+					    * $id none $date
+				}
+			}
 
 			return
 		} quiet {
@@ -1190,7 +1189,13 @@ proc ::irctk::writemsg {chan id tags text {loop 0}} {
 		}
 	}
 
-	irc send $chan $tags $cmd [::join $args " "]
+	switch -exact -- [irc send $chan $tags $cmd [::join $args " "]] {
+		-2 {
+			gui displaymsg "message too long" * $id none $date
+		} -1 {
+			gui displaymsg "can not send message" * $id none $date
+		}
+	}
 }
 
 proc ::irctk::donotdisturb {{text ""}} {
@@ -1233,16 +1238,17 @@ proc ::irctk::findnetwork {id} {
 #
 # Save for the next time.
 #
-proc ::irctk::save {network nick name pass cmd} {
+proc ::irctk::save {network user nick name pass cmd} {
 	variable servers
 	variable newservers
 
 	dict set newservers network $network name $name
 	dict set newservers network $network nick $nick
-	dict set newservers network $network pass $pass
+	dict set newservers network $network pass "$pass"
+	dict set newservers network $network user "$user"
 	dict set newservers network $network cmd "$cmd"
 
-	set params [::list $network $nick $name $pass "$cmd"]
+	set params [::list $network $nick $name "$pass" "$cmd" "$user"]
 
 	set cfgdir "$::cfgdir/auto/$network"
 
@@ -1293,6 +1299,8 @@ proc ::irctk::quitserver {chan} {
 		log console "$errstr"
 	}
 
+	gui closechannel [dict get $servers $chan channels server id]
+
 	dict unset servers $chan
 
 	#
@@ -1327,17 +1335,76 @@ proc ::irctk::numericresp {chan} {
 		001 {
 			set id [dict get $servers $chan channels server id]
 
+			set givennick [lindex [dict get $msg args] 0]
+			set nick [dict get $servers $chan nick]
+			set network [dict get $servers $chan network]
+			set user [dict get $servers $chan user]
+			set name [dict get $servers $chan name]
+			set pass [dict get $servers $chan pass]
+			set cmd [dict get $servers $chan cmd]
+
+			if {"$givennick" ne $nick} {
+				save $network "$user" $nick $name "$pass" "$cmd"
+
+				dict set servers $chan nick $givennick
+			}
+
+			gui updateuser $id $nick "$givennick"
+
 			gui displaymsg [join [dict get $msg args] " "]  "*" \
 			    $id info $date
 
 			autojoin $chan
 
 			after 8000 "::irctk::setok $chan"
-		} 00[2-5] {
+		} 00[2-4] {
 			set id [dict get $servers $chan channels server id]
 
 			gui displaymsg [join [dict get $msg args] " "]  "*" \
 			    $id info $date
+		} 005 {
+			set id [dict get $servers $chan channels server id]
+
+			gui displaymsg [join [dict get $msg args] " "]  "*" \
+			    $id info $date
+
+			foreach arg [lreplace [dict get $msg args] 0 0] {
+				#
+				# The last argument in the message is
+				# just a comment.
+				#
+				# It means we are done processing.
+				#
+				if {[string index $arg 0] eq ":"} {break}
+
+				#
+				# Features are in the form PARAM=VALUE
+				#
+				set feat [split $arg "="]
+				set fname [lindex $feat 0]
+
+				#
+				# If a feature name has a "-" in front, we
+				# must remove it.
+				#
+				if {[string index $fname 0] eq "-"} {
+					set fname [string trimleft $fname "-"]
+					dict remove servers $chan features $fname
+					continue
+				}
+
+				set values {}
+
+				#
+				# A feature may have multiple values, those
+				# are separated by a comma.
+				#
+				foreach val [split [lindex $feat 1] ","] {
+					lappend values $val
+				}
+
+				dict set servers $chan features $fname values $values
+			}
 		} 305 {
 			dict set servers $chan away 0
 
@@ -1372,6 +1439,13 @@ proc ::irctk::numericresp {chan} {
 				set users [lreplace $users 0 0 [string trimleft \
 				    [lindex $users 0] ":"]]
 
+				if {[dict get $servers $chan channels $channel \
+				    names] == 0} {
+					dict set servers $chan channels \
+					    $channel names 1
+					gui clearusers $id
+				}
+
 				foreach user $users {
 					gui adduser $id $user
 				}
@@ -1379,7 +1453,13 @@ proc ::irctk::numericresp {chan} {
 				gui displaymsg [join [dict get $msg args] \
 				    " "] "*" -1 info $date
 			}
-		} 3(28|66) {
+		} 366 {
+			set channel [lindex [dict get $msg args] 1]
+
+			if {[dict exists $servers $chan channels $channel]} {
+				dict set servers $chan channels $channel names 0
+			}
+		} 328 {
 		} 333 {
 			if {[dict get $servers $chan ok]} {
 				gui displaymsg [join [dict get $msg args] \
@@ -1406,7 +1486,11 @@ proc ::irctk::ischanrd {chan channel} {
 		return 0
 	}
 
-	return [dict get $servers $chan channels $channel readonly]
+	if {[dict exists $servers $chan channels $channel readonly]} {
+		return [dict get $servers $chan channels $channel readonly]
+	}
+
+	return 0
 }
 
 #
@@ -1416,7 +1500,7 @@ proc ::irctk::ischanrd {chan channel} {
 # If an extension sends a message for a channel that does not
 # exists, we create a "read only" one.
 #
-# Useful to just show the informations coming from the extension
+# Useful to just show the information coming from the extension
 # itself. We can kill and close it like any other, but we are not
 # allowed to write messages on it from the GUI.
 #
@@ -1431,7 +1515,7 @@ proc ::irctk::extmsg {msg} {
 
 	set channel [dict get $msg channel]
 	set network [dict get $msg network]
-	set msgstr [concat [dict get $msg args]]
+	set msgstr [join [dict get $msg args] " "]
 	set ch [findchan $network $channel]
 
 	if {[llength $ch] == 0} {
